include string.e


-- Read a {}-enclosed list of numbers separated by spaces or commas. The list can
-- contain a loop, beginning at | and ranging to the end of the list.
global function get_list()
	sequence s,t
	integer c, commaOk, endOk, pipeOk, concatTo
	object o
	
	-- s[2] will contain everything before the |, s[3] everything after it.
	s = {GET_FAIL, {}, {}}
	
	commaOk = 0	-- Not ok to read a comma
	pipeOk = 1	-- Ok to read a |
	endOk = 0	-- Not ok to read a }
	concatTo = 2	-- Concatenate to s[2]
	
	skip_whitespace()
	
	c = getch()
	
	if c = '{' then
		while 1 do
			skip_whitespace()
			t = get_string_in_range(DIGITS & '-')
			if length(t) then
				o = value(t)
				if o[1] = GET_SUCCESS then
					s[concatTo] &= o[2]
					commaOk = 1
					pipeOk = 1
					endOk = 1
				else
					ERROR("Syntax error: " & t, lineNum)
				end if
			else
				c = getch()
				if c = ',' then
					if not commaOk then
						ERROR("Unexpected comma", lineNum)
					end if
					commaOk = 0
					pipeOk = 0
					endOk = 0
				elsif c = '|' then
					if concatTo = 2 and pipeOk then
						concatTo = 3
						commaOk = 0
						pipeOk = 0
						endOk = 0
					else
						ERROR("Unexpected |", lineNum)
					end if
				elsif c = '}' then
					if endOk then
						s[1] = GET_SUCCESS
						exit
					else
						ERROR("Malformed list", lineNum)
					end if
				elsif c = ';' then
					while c != 10 and c != -1 do
						c = getch()
					end while
				elsif c = '\"' then
					t = ""
					c = getch()
					while c != '\"' and c != -1 do
						t &= c
						c = getch()
					end while
					s[concatTo] &= {t}
				elsif c = -1 then
					exit
				end if
			end if
		end while
	else
		ERROR("Expected {, got " & c, lineNum)
	end if
	
	return s
end function


-- Step through a list
-- Return {new_position next_item}
global function step_list(sequence l, sequence pos)
	sequence t
	if pos[1] <= length(l[pos[2]]) then
		t = {l[pos[2]][pos[1]]}
		pos[1] += 1
		t = pos[1..2] & t
	else 
		if length(l[3]) then
			pos[2] = 3
			pos[1] = 1
			t = {l[pos[2]][pos[1]]}
			pos[1] += 1
			t = pos[1..2] & t
		else
			t = {length(l[2]), 2, l[2][length(l[2])]}
		end if
	end if
	
	return t
end function
		
	
-- Return a sequence containing a text-form representation on a list returned by get_list
-- 
-- Examples:
--   {GET_SUCCESS, {1, 2, 3}, {}} -> "{1 2 3}"
--   {GET_SUCCESS, {1, 2, 3}, {4, 5}} -> "{1 2 3 | 4 5}"
--   {GET_SUCCESS, {}, {4, 5}} -> "{| 4 5}"
--
global function sprint_list(sequence l)
	sequence t
	
	t = "{"
	for i = 1 to length(l[2]) do
		if sequence(l[2][i]) then
			t &= l[2][i] & " "
		else
			t &= sprintf("%d", l[2][i])
		end if
		if i < length(l[2]) then
			t &= ' '
		end if
	end for
	
	if length(l[3]) then
		t &= " | "
		for i = 1 to length(l[3]) do
			t &= sprintf("%d", l[3][i])
			if i < length(l[3]) then
				t &= ' '
			end if
		end for
	end if
	
	return t & '}'
end function


global function is_empty_list(sequence l)
	return (length(l[2]) + length(l[3])) = 0
end function
